import os
from typing import List, Optional, Sequence, Tuple, Union

import plotly.graph_objects as go
import torch
from torch import Tensor, nn

from yet_another_retnet.retnet import RetNet, retnet_1_3b
from yet_another_retnet.utils.benchmark import benchmark
from yet_another_retnet.utils.profile import profile

NUM_TOKENS = 10000
BATCH_SIZE = 4
SEQ_LENGTHS = [2048, 3072, 4096, 5120, 6144, 7168, 8192]
if torch.cuda.is_available():
    DEVICE = torch.device("cuda")
    DTYPE = torch.float16
else:
    DEVICE = torch.device("cpu")
    DTYPE = torch.float32


class TransformerLM(nn.Module):
    def __init__(
        self,
        num_tokens: int,  # usually obtained from the tokenizer
        d_model: int,
        nhead: int,
        num_layers: int,
        dim_feedforward: int,
        max_batch_size: int = BATCH_SIZE,
        max_seq_length: int = 8192,
        device: Optional[Union[torch.device, str]] = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        super().__init__()
        self.embeddings = nn.Embedding(num_tokens, d_model, device=device, dtype=dtype)
        decoder_layer = nn.TransformerDecoderLayer(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            device=device,
            dtype=dtype,
        )
        self.decoder = nn.TransformerDecoder(
            decoder_layer=decoder_layer, num_layers=num_layers
        )
        self.out = nn.Linear(d_model, num_tokens, device=device, dtype=dtype)

        # TODO: This may not be identical to what was benchmarked in the paper.
        # They specifically mention a KV cache, and this isn't technically that.
        # (The key/value projections aren't being cached, just the memory values
        # before KV projection.)  However, this seemed easier to implement, since
        # I don't have to fiddle with Flash Attention or custom Transformer layer
        # implementations.  If time allows, consider implementing a KV cache.
        #
        # a rough-and-dirty memory (KV) cache
        self.cache = torch.zeros(
            (max_batch_size, max_seq_length, d_model),
            device=device,
            dtype=dtype,
        )

    def forward(self, x: Tensor, start_pos: int) -> Tensor:
        batch_size, seq_len = x.shape
        x = self.embeddings(x)

        # memory cache
        self.cache[:batch_size, start_pos : start_pos + seq_len] = x
        memory = self.cache[:batch_size, : start_pos + seq_len]
        x = self.decoder.forward(x, memory)
        return self.out(x)


def transformer_1_3b(
    num_tokens: int,  # usually obtained from the tokenizer
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> TransformerLM:
    """Transformer configuration to match RetNet 1.3B from the paper:
    https://arxiv.org/pdf/2307.08621v3.pdf
    """
    return TransformerLM(
        num_tokens=num_tokens,
        d_model=2048,
        nhead=8,
        num_layers=24,
        dim_feedforward=4096,
        device=device,
        dtype=dtype,
    )


@torch.no_grad()
def benchmark_inference_throughput(
    retnet: RetNet, transformer: TransformerLM, seq_lengths: Sequence[int]
) -> Tuple[List[float], List[float]]:
    retnet_throughputs: List[float] = []
    transformer_throughputs: List[float] = []

    print("\nBenchmarking inference throughput...")
    for seq_length in seq_lengths:
        torch.cuda.empty_cache()
        print(f"seq_length: {seq_length}")
        x = torch.randint(0, NUM_TOKENS, (BATCH_SIZE, seq_length), device=DEVICE)

        # Benchmark *recurrent* RetNet formulation for inference
        retnet_result = benchmark(
            retnet.forward_recurrent, x[:, 0], seq_idx=0, prev_states=[]
        )
        retnet_throughput = BATCH_SIZE / retnet_result.mean
        print(f"RetNet throughput: {retnet_throughput:.3f} tokens/s")
        # Benchmark *parallel* transformer for inference (with memory cache)
        _ = transformer(x, start_pos=0)  # warmup memory cache
        transformer_result = benchmark(transformer, x[:, -1:], start_pos=seq_length - 1)
        transformer_throughput = BATCH_SIZE / transformer_result.mean
        print(f"Transformer throughput: {transformer_throughput:.3f} tokens/s")

        retnet_throughputs.append(retnet_throughput)
        transformer_throughputs.append(transformer_throughput)

    return retnet_throughputs, transformer_throughputs


@torch.no_grad()
def measure_inference_memory(
    retnet: RetNet, transformer: TransformerLM, seq_lengths: Sequence[int]
) -> Tuple[List[float], List[float]]:
    retnet_memories: List[float] = []
    transformer_memories: List[float] = []

    print("\nMeasuring inference memory...")
    for seq_length in seq_lengths:
        torch.cuda.empty_cache()
        print(f"seq_length: {seq_length}")
        x = torch.randint(0, NUM_TOKENS, (BATCH_SIZE, seq_length), device=DEVICE)

        # Measure *recurrent* RetNet formulation for inference
        retnet_result = profile(
            retnet.forward_recurrent, x[:, 0], seq_idx=0, prev_states=[]
        )
        retnet_memory_gib = retnet_result.peak / 2**30
        print(f"RetNet memory: {retnet_memory_gib:.3f} GiB")
        # Measure *parallel* transformer for inference (with memory cache)
        _ = transformer(x, start_pos=0)  # warmup memory cache
        transformer_result = profile(transformer, x[:, -1:], start_pos=seq_length - 1)
        transformer_memory_gib = transformer_result.peak / 2**30
        print(f"Transformer memory: {transformer_memory_gib:.3f} GiB")

        retnet_memories.append(retnet_memory_gib)
        transformer_memories.append(transformer_memory_gib)

    return retnet_memories, transformer_memories


if __name__ == "__main__":
    retnet = retnet_1_3b(NUM_TOKENS, device=DEVICE, dtype=DTYPE).eval()
    transformer = transformer_1_3b(NUM_TOKENS, device=DEVICE, dtype=DTYPE).eval()

    if torch.cuda.is_available():
        retnet_footprints, transformer_footprints = measure_inference_memory(
            retnet, transformer, seq_lengths=SEQ_LENGTHS
        )
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=SEQ_LENGTHS,
                y=retnet_footprints,
                name="RetNet",
                mode="lines+markers",
                line={"color": "blue"},
                marker={"color": "blue"},
            )
        )
        fig.add_trace(
            go.Scatter(
                x=SEQ_LENGTHS,
                y=transformer_footprints,
                name="Transformer",
                mode="lines+markers",
                line={"color": "red"},
                marker={"color": "red"},
            )
        )
        fig.update_layout(
            title="Inference Memory Footprint",
            xaxis_title="Sequence Length",
            yaxis_title="GPU Memory (GiB)",
            xaxis={"tickmode": "array", "tickvals": SEQ_LENGTHS},
            # place legend at center-left
            legend={"x": 0.1, "y": 0.5},
        )
        fig.write_image(os.path.join("doc", "inference-memory.png"))
    else:
        print("Skipping GPU memory profiling, because CUDA is not available.")

    retnet_throughputs, transformer_throughputs = benchmark_inference_throughput(
        retnet, transformer, seq_lengths=SEQ_LENGTHS
    )
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=SEQ_LENGTHS,
            y=retnet_throughputs,
            name="RetNet",
            mode="lines+markers",
            line={"color": "blue"},
            marker={"color": "blue"},
        )
    )
    fig.add_trace(
        go.Scatter(
            x=SEQ_LENGTHS,
            y=transformer_throughputs,
            name="Transformer",
            mode="lines+markers",
            line={"color": "red"},
            marker={"color": "red"},
        )
    )
    fig.update_layout(
        title="Inference Throughput",
        xaxis_title="Sequence Length",
        yaxis_title="Throughput (tokens/s)",
        xaxis={"tickmode": "array", "tickvals": SEQ_LENGTHS},
        # place legend at center-left
        legend={"x": 0.1, "y": 0.5},
    )
    fig.write_image(os.path.join("doc", "inference-throughput.png"))



import pytest


def pytest_addoption(parser):
    parser.addoption("--slow", action="store_true")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: slow to run")


def pytest_collection_modifyitems(config, items):
    run_slow = config.getoption("--slow")
    skip_fast = pytest.mark.skip(reason="remove --slow option to run")
    skip_slow = pytest.mark.skip(reason="need --slow option to run")

    for item in items:
        if ("slow" in item.keywords) and (not run_slow):
            item.add_marker(skip_slow)
        if ("slow" not in item.keywords) and (run_slow):
            item.add_marker(skip_fast)



from typing import Optional, Sequence

import pytest
import torch
from torch import Tensor

from yet_another_retnet.retnet import RetNet

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float32
# Set deterministic CUDA ops
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False


def test_retnet_forward_parallel():
    # TODO
    pass


def test_retnet_forward_recurrent():
    # TODO
    pass


@torch.no_grad()
@pytest.mark.parametrize("num_tokens", [10, 100])
@pytest.mark.parametrize("d_model", [16, 32])
@pytest.mark.parametrize("nhead", [1, 2])
@pytest.mark.parametrize("num_layers", [1, 2])
@pytest.mark.parametrize("batch_size", [2])
@pytest.mark.parametrize("seq_length", [8])
@pytest.mark.parametrize("chunk_size", [1, 4, 8])
def test_retnet_equivalent_formulations(
    num_tokens: int,
    d_model: int,
    nhead: int,
    num_layers: int,
    batch_size: int,
    seq_length: int,
    chunk_size: int,
):
    size = (batch_size, seq_length)
    x = torch.randint(0, num_tokens, size=size, device=DEVICE)
    net = RetNet(
        num_tokens=num_tokens,
        d_model=d_model,
        nhead=nhead,
        num_layers=num_layers,
        device=DEVICE,
        dtype=DTYPE,
    ).eval()

    y_parallel = net.forward_parallel(x)

    y_recurrent = torch.zeros_like(y_parallel)
    prev_states: Sequence[Optional[Tensor]] = [None] * num_layers
    for i in range(seq_length):
        xr = x[:, i]
        y_recurrent[:, i], prev_states = net.forward_recurrent(
            xr, seq_idx=i, prev_states=prev_states
        )
    recurrent_states = prev_states

    torch.testing.assert_close(y_parallel, y_recurrent, rtol=1e-4, atol=1e-4)

    y_chunkwise = torch.zeros_like(y_parallel)
    prev_states = [None] * num_layers
    for i in range(0, seq_length, chunk_size):
        xc = x[:, i : i + chunk_size]
        y_chunkwise[:, i : i + chunk_size], prev_states = net.forward_chunkwise(
            xc, start_idx=i, prev_states=prev_states
        )

    torch.testing.assert_close(y_parallel, y_chunkwise, rtol=1e-4, atol=1e-4)
    assert len(recurrent_states) == len(prev_states)
    for rstate, cstate in zip(recurrent_states, prev_states):
        torch.testing.assert_close(rstate, cstate, rtol=1e-4, atol=1e-4)



from copy import deepcopy
from typing import Callable, List, Optional, Sequence, Tuple, Union

import torch
from einops import rearrange
from torch import Tensor, nn

from yet_another_retnet.retention import (
    ActivationString,
    MultiScaleRetention,
    _get_activation_fn,
)


class RetNetDecoderLayer(nn.Module):
    # NOTE: Mostly pulled from 'nn.TransformerDecoderLayer', but with changes:
    #   - use MultiScaleRetention instead of MultiheadAttention
    #   - no cross-attention layer, since retention doesn't play well with that

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[ActivationString, Callable[[Tensor], Tensor]] = "swish",
        norm_first: bool = True,
        layer_norm_eps: float = 1e-6,
        device: Optional[Union[torch.device, str]] = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)

        super().__init__()
        self.dropout = nn.Dropout(dropout)
        self.activation = activation
        self.norm_first = norm_first
        # retention block
        self.norm1 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.retention = MultiScaleRetention(  # type: ignore
            embed_dim=d_model,
            num_heads=nhead,
            dropout=dropout,
            activation=activation,
            device=device,
            dtype=dtype,
        )
        # feedforward block
        self.norm2 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.linear1 = nn.Linear(d_model, dim_feedforward, device=device, dtype=dtype)
        self.linear2 = nn.Linear(dim_feedforward, d_model, device=device, dtype=dtype)

        self._reset_parameters()

    def _reset_parameters(self):
        # TODO: Check that we're following the same initialization as the paper
        nn.init.xavier_normal_(self.linear1.weight)
        nn.init.constant_(self.linear1.bias, 0)
        nn.init.xavier_normal_(self.linear2.weight)
        nn.init.constant_(self.linear2.bias, 0)

    def _feedforward_block(self, x: Tensor) -> Tensor:
        x = self.activation(self.linear1(x))
        x = self.dropout(x)
        x = self.linear2(x)
        x = self.dropout(x)
        return x

    def forward_parallel(self, x: Tensor) -> Tensor:
        def _retention_block(x: Tensor) -> Tensor:
            x, _ = self.retention.forward_parallel(x, x, x)
            return self.dropout(x)

        if self.norm_first:
            x = x + _retention_block(self.norm1(x))
            x = x + self._feedforward_block(self.norm2(x))
        else:
            x = x + self.norm1(_retention_block(x))
            x = x + self.norm2(self._feedforward_block(x))

        return x

    def forward_recurrent(
        self, x: Tensor, seq_idx: int, prev_state: Optional[Tensor] = None
    ) -> Tuple[Tensor, Tensor]:
        def _retention_block(x: Tensor) -> Tuple[Tensor, Tensor]:
            x, state = self.retention.forward_recurrent(
                x, x, x, seq_idx=seq_idx, prev_state=prev_state
            )
            return self.dropout(x), state

        # retention block
        if self.norm_first:
            y, state = _retention_block(self.norm1(x))
            x = x + y
            x = x + self._feedforward_block(self.norm2(x))
        else:
            y, state = _retention_block(x)
            x = x + self.norm1(y)
            x = x + self.norm2(self._feedforward_block(x))

        return x, state

    def forward_chunkwise(
        self, x: Tensor, start_idx: int, prev_state: Optional[Tensor] = None
    ) -> Tuple[Tensor, Tensor]:
        def _retention_block(x: Tensor) -> Tuple[Tensor, Tensor]:
            x, state = self.retention.forward_chunkwise(
                x, x, x, start_idx=start_idx, prev_state=prev_state
            )
            return self.dropout(x), state

        # retention block
        if self.norm_first:
            y, state = _retention_block(self.norm1(x))
            x = x + y
            x = x + self._feedforward_block(self.norm2(x))
        else:
            y, state = _retention_block(x)
            x = x + self.norm1(y)
            x = x + self.norm2(self._feedforward_block(x))

        return x, state

    def forward(self, x: Tensor) -> Tensor:
        return self.forward_parallel(x)


class RetNetDecoder(nn.Module):
    def __init__(self, decoder_layer: RetNetDecoderLayer, num_layers: int):
        super().__init__()
        self.num_layers = num_layers
        self.layers = nn.ModuleList(
            [deepcopy(decoder_layer) for _ in range(num_layers)]
        )

    def forward_parallel(self, x: Tensor) -> Tensor:
        for layer in self.layers:
            assert isinstance(layer, RetNetDecoderLayer)
            x = layer.forward_parallel(x)
        return x

    def forward_recurrent(
        self, x: Tensor, seq_idx: int, prev_states: Sequence[Optional[Tensor]] = ()
    ) -> Tuple[Tensor, List[Tensor]]:
        if not prev_states:
            prev_states = [None] * self.num_layers
        elif len(prev_states) != len(self.layers):
            raise ValueError(
                f"Expected {len(self.layers)} previous states, got {len(prev_states)}"
            )

        states: List[Tensor] = []
        for layer, prev_state in zip(self.layers, prev_states):
            assert isinstance(layer, RetNetDecoderLayer)
            x, state = layer.forward_recurrent(x, seq_idx, prev_state)
            states.append(state)
        return x, states

    def forward_chunkwise(
        self, x: Tensor, start_idx: int, prev_states: Sequence[Optional[Tensor]] = ()
    ) -> Tuple[Tensor, List[Tensor]]:
        if not prev_states:
            prev_states = [None] * self.num_layers
        elif len(prev_states) != len(self.layers):
            raise ValueError(
                f"Expected {len(self.layers)} previous states, got {len(prev_states)}"
            )

        states: List[Tensor] = []
        for layer, prev_state in zip(self.layers, prev_states):
            assert isinstance(layer, RetNetDecoderLayer)
            x, state = layer.forward_chunkwise(x, start_idx, prev_state)
            states.append(state)
        return x, states

    def forward(self, x: Tensor) -> Tensor:
        return self.forward_parallel(x)


class RetNet(nn.Module):
    def __init__(
        self,
        num_tokens: int,  # usually obtained from the tokenizer
        d_model: int = 512,
        nhead: int = 8,
        num_layers: int = 6,
        dropout: float = 0.1,
        activation: Union[ActivationString, Callable[[Tensor], Tensor]] = "swish",
        dim_feedforward: int = 2048,
        norm_first: bool = True,
        layer_norm_eps: float = 1e-6,
        device: Optional[Union[torch.device, str]] = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        super().__init__()
        self.d_model = d_model
        self.num_layers = num_layers
        self.embedding = nn.Embedding(num_tokens, d_model, device=device, dtype=dtype)
        decoder_layer = RetNetDecoderLayer(
            d_model,
            nhead,
            dropout=dropout,
            activation=activation,
            dim_feedforward=dim_feedforward,
            norm_first=norm_first,
            layer_norm_eps=layer_norm_eps,
            device=device,
            dtype=dtype,
        )
        self.decoder = RetNetDecoder(decoder_layer, num_layers)
        self.out = nn.Linear(d_model, num_tokens, device=device, dtype=dtype)

        self._reset_parameters()

    def _reset_parameters(self):
        nn.init.xavier_normal_(self.out.weight)
        nn.init.constant_(self.out.bias, 0)

    def forward_parallel(self, x: Tensor) -> Tensor:
        x = self.embedding(x)
        x = self.decoder.forward_parallel(x)
        x = self.out(x)
        return x

    def forward_recurrent(
        self, x: Tensor, seq_idx: int, prev_states: Sequence[Optional[Tensor]] = ()
    ) -> Tuple[Tensor, List[Tensor]]:
        x = self.embedding(x)
        x, states = self.decoder.forward_recurrent(
            x, seq_idx=seq_idx, prev_states=prev_states
        )
        x = self.out(x)
        return x, states

    def forward_chunkwise(
        self, x: Tensor, start_idx: int, prev_states: Sequence[Optional[Tensor]] = ()
    ) -> Tuple[Tensor, List[Tensor]]:
        x = self.embedding(x)
        x, states = self.decoder.forward_chunkwise(
            x, start_idx=start_idx, prev_states=prev_states
        )
        x = self.out(x)
        return x, states

    def forward(self, inputs: Tensor, labels: Tensor) -> Tensor:
        pred = self.forward_parallel(inputs)
        criterion = nn.CrossEntropyLoss()
        return criterion(rearrange(pred, "b n c -> (b n) c"), labels.flatten())


def retnet_1_3b(
    num_tokens: int,  # usually obtained from the tokenizer
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> RetNet:
    """RetNet 1.3B configuration from the paper:
    https://arxiv.org/pdf/2307.08621v3.pdf
    """
    return RetNet(
        num_tokens=num_tokens,
        d_model=2048,
        nhead=8,
        num_layers=24,
        dim_feedforward=4096,
        device=device,
        dtype=dtype,
    )


def retnet_2_7b(
    num_tokens: int,  # usually obtained from the tokenizer
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> RetNet:
    """RetNet 2.7B configuration from the paper:
    https://arxiv.org/pdf/2307.08621v3.pdf
    """
    return RetNet(
        num_tokens=num_tokens,
        d_model=2560,
        nhead=10,
        num_layers=32,
        dim_feedforward=5120,
        device=device,
        dtype=dtype,
    )


def retnet_6_7b(
    num_tokens: int,  # usually obtained from the tokenizer
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> RetNet:
    """RetNet 6.7B configuration from the paper:
    https://arxiv.org/pdf/2307.08621v3.pdf
    """
    return RetNet(
        num_tokens=num_tokens,
        d_model=4096,
        nhead=16,
        num_layers=32,
        dim_feedforward=8192,
        device=device,
        dtype=dtype,
    )


if __name__ == "__main__":
    num_tokens = 1000
    batch_size = 1
    seq_len = 8
    d_model = 32
    nhead = 2
    num_layers = 2
    device = "cuda"
    dtype = torch.float32

    size = (batch_size, seq_len)
    x = torch.randint(0, num_tokens, size=size, device=device)
    net = RetNet(
        num_tokens=num_tokens,
        d_model=d_model,
        nhead=nhead,
        num_layers=num_layers,
        device=device,
        dtype=dtype,
    ).eval()

    y_parallel = net.forward_parallel(x)

    y_recurrent = torch.zeros_like(y_parallel)
    prev_states: Sequence[Optional[Tensor]] = [None] * num_layers
    for i in range(seq_len):
        xr = x[:, i]
        y_recurrent[:, i], prev_states = net.forward_recurrent(
            xr, seq_idx=i, prev_states=prev_states
        )

    torch.testing.assert_close(y_parallel, y_recurrent)



import os
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import tiktoken
import torch
from lightning import Fabric, seed_everything
from lightning.fabric.loggers import TensorBoardLogger
from torch import Tensor
from torch.utils.data import DataLoader
from tqdm import tqdm

from yet_another_retnet.retnet import RetNet
from yet_another_retnet.utils.gutenberg import project_gutenberg_top_100_datapipe

torch.set_float32_matmul_precision("medium")
TOKENIZER = tiktoken.get_encoding("gpt2")
EVAL_PROMPT = "A Lannister always pays his debts."


def collate_fn(
    batch: List[str],
    max_length: int = 1024,
    device: Optional[Union[torch.device, str]] = None,
) -> Tuple[Tensor, Tensor]:
    x = torch.zeros(len(batch), max_length, device=device, dtype=torch.long)
    y = torch.zeros(len(batch), max_length, device=device, dtype=torch.long)

    for i, text in enumerate(batch):
        encoding = torch.as_tensor(
            TOKENIZER.encode(text), device=device, dtype=torch.long
        )
        seq_length = min(len(encoding) - 1, max_length)
        x[i, :seq_length] = encoding[:seq_length]
        y[i, :seq_length] = encoding[1 : seq_length + 1]

    return x, y


@dataclass
class TrainingState:
    fabric: Fabric
    model: RetNet
    optimizer: torch.optim.Optimizer
    callbacks: Sequence[Callable[["TrainingState", float], None]] = ()

    current_step: int = 0
    current_epoch: int = 0
    accumulate_grad_batches: int = 1
    monitor: str = "val_loss"
    monitor_mode: Literal["min", "max"] = "min"


@dataclass
class ModelCheckpoint:
    state_dict: Dict[str, Tensor]
    optimizer_state: Dict[str, Tensor]
    current_step: int
    current_epoch: int

    @classmethod
    def from_training_state(cls, state: TrainingState) -> "ModelCheckpoint":
        return cls(
            state_dict=state.model.state_dict(),
            optimizer_state=state.optimizer.state_dict(),
            current_step=state.current_step,
            current_epoch=state.current_epoch,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "state_dict": self.state_dict,
            "optimizer_state": self.optimizer_state,
            "current_step": self.current_step,
            "current_epoch": self.current_epoch,
        }

    def save(self, path: str) -> None:
        torch.save(self.to_dict(), path)

    @classmethod
    def load(cls, path: str) -> "ModelCheckpoint":
        checkpoint_dict = torch.load(path)
        return cls(**checkpoint_dict)


class CheckpointCallback:
    def __init__(
        self, save_dir: str, name: str = "checkpoint_epoch-{epoch:03d}.pt"
    ) -> None:
        self.save_dir = save_dir
        self.name = name
        self.best_path: Optional[str] = None
        self.best_loss: Optional[float] = None

    def __call__(self, state: TrainingState, loss: float) -> None:
        if self.best_loss is None:
            self.best_loss = loss

        fabric = state.fabric
        # 'local_rank == 0' means this only happens for the main process
        if fabric.local_rank == 0 and loss <= self.best_loss:
            checkpoint = ModelCheckpoint.from_training_state(state)
            self.best_loss = loss
            if self.best_path is not None:
                os.remove(self.best_path)
            self.best_path = os.path.join(
                self.save_dir, self.name.format(epoch=state.current_epoch)
            )
            torch.save(checkpoint, self.best_path)

        # All processes wait for main to finish saving the checkpoint.
        fabric.barrier()


def train_one_epoch(
    state: TrainingState,
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    log_frequency: int = 25,
) -> None:
    state.current_epoch += 1
    fabric, model, optimizer = state.fabric, state.model, state.optimizer
    is_main_process = fabric.local_rank == 0
    is_training = model.training
    model.train()

    with tqdm(
        desc=f"Ep: {state.current_epoch}", disable=(not is_main_process)
    ) as progbar:
        train_loss, val_loss = 0.0, 0.0
        for x, y in train_dataloader:
            state.current_step += 1
            accumulating = state.current_step % state.accumulate_grad_batches != 0
            with fabric.no_backward_sync(model, enabled=accumulating):  # type: ignore
                loss = model.forward(inputs=x, labels=y)
                fabric.backward(loss)

            if not accumulating:
                optimizer.step()
                optimizer.zero_grad()

            if state.current_step % log_frequency == 0:
                fabric.log("loss", loss, step=state.current_step)
                train_loss = loss.item()
                progbar.set_postfix_str(f"loss={train_loss:.4f}", refresh=False)
            progbar.update(1)

        model.eval()
        val_progbar = tqdm(desc="val", position=1, leave=False)
        for i, (x, y) in enumerate(val_dataloader):
            with torch.inference_mode():
                loss = model.forward(inputs=x, labels=y)
            val_loss = (val_loss * i + loss.item()) / (i + 1)

            if i % log_frequency == 0:
                val_progbar.set_postfix_str(f"val_loss={val_loss:.4f}", refresh=False)
            val_progbar.update(1)
            progbar.update(1)

        fabric.log("val_loss", val_loss, step=state.current_step)
        val_progbar.close()
        progbar.set_postfix_str(
            f"loss={train_loss:.4f}, val_loss={val_loss:.4f}", refresh=False
        )

        for callback in state.callbacks:
            callback(state, val_loss)

        # Return model to its original training state
        model.train(mode=is_training)


def train(
    retnet: RetNet,
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    accelerator: str = "auto",
    strategy: str = "auto",
    precision: Optional[str] = None,
    epochs: int = 10,
    lr: float = 3e-4,
    log_frequency: int = 25,
):
    if precision is None:
        if torch.cuda.is_available():
            # use bfloat16 if supported
            version, _ = torch.cuda.get_device_capability()
            precision = "bf16-mixed" if version >= 8 else "16-mixed"
        else:
            precision = "32-true"

    logger = TensorBoardLogger(root_dir="./")
    fabric = Fabric(
        accelerator=accelerator,
        strategy=strategy,
        precision=precision,  # type: ignore
        loggers=[logger],
    )
    fabric.launch()
    print(f"Experiment version: {logger.version}")
    print("-" * 40)

    # Setup with fabric.
    optimizer = torch.optim.AdamW(retnet.parameters(), lr=lr)
    retnet, optimizer = fabric.setup(retnet, optimizer)
    train_dataloader, val_dataloader = fabric.setup_dataloaders(
        train_dataloader, val_dataloader
    )
    # Construct a training state and run the training loop.
    state = TrainingState(
        fabric=fabric,
        model=retnet,
        optimizer=optimizer,
        callbacks=[CheckpointCallback(save_dir=logger.log_dir)],
    )
    for _ in range(epochs):
        train_one_epoch(
            state=state,
            train_dataloader=train_dataloader,
            val_dataloader=val_dataloader,
            log_frequency=log_frequency,
        )


def generate(
    retnet: RetNet,
    prompt: str,
    prompt_chunk_size: Optional[int] = None,
    max_new_tokens: int = 4096,
    stop_tokens: Sequence[str] = (),
    top_k: int = 10,
    temperature: float = 1.0,
    seed: int = 42,
) -> Iterator[str]:
    seed_everything(seed)
    device = next(iter(retnet.parameters())).device
    is_training = retnet.training
    retnet.eval()

    # Tokenize the prompt and convert to a tensor.
    tokenized = TOKENIZER.encode(prompt)
    x = torch.as_tensor(tokenized, dtype=torch.long, device=device).unsqueeze_(0)

    if not prompt_chunk_size:
        prompt_chunk_size = x.size(1)

    prev_states: List[Optional[Tensor]] = [None] * retnet.num_layers
    start_idx: int = 0
    for start_idx in range(0, x.size(1), prompt_chunk_size):
        y, prev_states = retnet.forward_chunkwise(  # type: ignore
            x, start_idx=start_idx, prev_states=prev_states
        )
        y = y[:, -1]

    # Generate tokens until we reach the maximum number of tokens or a stop token.
    for i in range(max_new_tokens):
        probs: Tensor = torch.softmax(y.squeeze() / max(temperature, 1e-8), dim=-1)
        # Get top-k tokens, renormalize their probabilities, and weighted sample.
        tokens: Tensor  # for mypy
        probs, tokens = probs.topk(k=top_k, dim=-1)
        probs /= probs.sum()

        # Take weighted random sample from the top-k tokens.
        sampled_idx: int = torch.multinomial(probs, num_samples=1).item()  # type: ignore
        token: int = tokens[sampled_idx].item()  # type: ignore
        tokenized.append(token)
        yield TOKENIZER.decode(tokenized)

        token_str: str = TOKENIZER.decode([token])
        if token_str in stop_tokens:
            break
        elif i < (max_new_tokens - 1):
            start_idx += 1
            x = torch.as_tensor([token], dtype=torch.long, device=device)
            y, prev_states = retnet.forward_recurrent(  # type: ignore
                x, start_idx, prev_states=prev_states
            )

    # Restore the model's original training state.
    retnet.train(mode=is_training)


def main(
    model_checkpoint: Optional[str] = None,
    accelerator: str = "auto",
    strategy: str = "auto",
    precision: Optional[str] = None,
    epochs: int = 10,
    batch_size: int = 16,
    lr: float = 3e-4,
    log_frequency: int = 25,
    seed: int = 42,
    eval_only: bool = False,
    eval_prompt: str = EVAL_PROMPT,
    eval_max_tokens: int = 1024,
):
    seed_everything(seed)
    # Create a (relatively small) model and dataloaders
    retnet = RetNet(
        num_tokens=TOKENIZER.n_vocab,
        d_model=768,
        nhead=8,
        num_layers=12,
    )
    if model_checkpoint is not None:
        retnet.load_state_dict(ModelCheckpoint.load(model_checkpoint).state_dict)

    if not eval_only:
        num_devices = torch.cuda.device_count()
        if num_devices > 0:
            # Lightning Fabric does not scale the batch size for distributed training.
            # In order to keep batch size the same, divide by the number of devices.
            if batch_size % num_devices != 0:
                raise ValueError(f"{batch_size=} must be divisible by {num_devices=}.")
            batch_size = batch_size // num_devices

        train_dataloader = DataLoader(
            project_gutenberg_top_100_datapipe(
                split="train",
                chunk_size=4096,
                step_size=1024,
                shuffle=True,
                drop_last=True,
            ),
            batch_size=batch_size,
            collate_fn=collate_fn,
            drop_last=True,
        )
        val_dataloader = DataLoader(
            project_gutenberg_top_100_datapipe(
                split="val", chunk_size=4096, step_size=1024
            ),
            batch_size=batch_size,
            collate_fn=collate_fn,
        )

        train(
            retnet=retnet,
            train_dataloader=train_dataloader,
            val_dataloader=val_dataloader,
            accelerator=accelerator,
            strategy=strategy,
            precision=precision,
            epochs=epochs,
            lr=lr,
            log_frequency=log_frequency,
        )

    # Generate some text
    prev_output: str = ""
    for output in generate(retnet, eval_prompt, max_new_tokens=eval_max_tokens):
        # Return to the start of the line and print the output (no newline)
        print(output[len(prev_output) :], end="", flush=True)
        prev_output = output
    print()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--model-checkpoint", type=str, default=None)
    parser.add_argument("--accelerator", type=str, default="auto")
    parser.add_argument("--strategy", type=str, default="auto")
    parser.add_argument("--precision", type=str, default=None)
    parser.add_argument("--epochs", type=int, default=10)
    parser.add_argument("--batch-size", type=int, default=16)
    parser.add_argument("--lr", type=float, default=3e-4)
    parser.add_argument("--log-frequency", type=int, default=25)
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--eval-only", action="store_true")
    parser.add_argument("--eval-prompt", type=str, default=EVAL_PROMPT)
    parser.add_argument("--eval-max-tokens", type=int, default=1024)
    args = parser.parse_args()

    main(**vars(args))



from functools import lru_cache
from math import log
from typing import Callable, Literal, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from einops import einsum, rearrange, repeat
from torch import Tensor, nn

DEFAULT_DEVICE = torch.device("cpu")
ActivationString = Literal["swish", "gelu", "relu"]


def _get_activation_fn(activation: str) -> Callable[[Tensor], Tensor]:
    """Return an activation function given a string"""
    if activation == "swish":
        return F.silu
    elif activation == "gelu":
        return F.gelu
    elif activation == "relu":
        return F.relu
    else:
        raise RuntimeError(
            f"Unsupported activation string '{activation}'. "
            "Supported: 'swish', 'gelu', 'relu'"
        )


@lru_cache(maxsize=1)
def _build_decay_gammas(
    num_heads: int,
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """Decay values are different for each retention head, following the prescribed
    method in the paper.  Conceptually, I think of each head having a different
    "retention window", which is the effective number of steps back in time that
    the head can attend to.  Retention windows are effectively determined by
    these decay coefficients.

    See: https://arxiv.org/pdf/2307.08621v3.pdf, Section 3.1 (Setup)
    """
    xmin, xmax = log(1 / 32), log(1 / 512)
    x = torch.linspace(xmin, xmax, steps=num_heads, device=device, dtype=dtype)
    return 1 - x.exp_()


@lru_cache(maxsize=1)
def _build_decay_mask(
    num_heads: int,
    query_length: int,
    key_length: int,
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """The decay mask is one of the key components that makes *parallel* retention
    equivalent to *recurrent* retention.  The decay coefficients are pre-computed
    and applied to the similarity matrix at once, rather than being applied to
    each element in the recurrent formulation.

    See: https://arxiv.org/pdf/2307.08621v3.pdf, Equation 5
    """
    decay_gammas = _build_decay_gammas(num_heads=num_heads, device=device, dtype=dtype)

    query_pos = torch.arange(query_length, device=device, dtype=dtype).unsqueeze_(-1)
    key_pos = torch.arange(key_length, device=device, dtype=dtype).unsqueeze_(0)
    distance = torch.abs(query_pos - key_pos)
    # Set the upper-triangular distances to infinity, so that only *past* keys
    # can affect the current query.  (Setting distance to infinity ensures that
    # the decay matrix is 0 for those positions, since x^(inf) = 0 when -1 < x < 1.
    distance_mask = torch.ones_like(distance, dtype=torch.bool).triu_(diagonal=1)
    distance = distance.masked_fill_(distance_mask, float("inf"))

    distance = rearrange(distance, "n s -> () n s")
    decay_gammas = rearrange(decay_gammas, "h -> h () ()")
    return decay_gammas**distance


def _build_position_thetas(
    head_dim: int,
    scale: float = 10000,
    device: Optional[Union[torch.device, str]] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """Positional thetas are different for each value along head_dim, following the
    prescribed method in the paper.  These are used to update the positional
    embeddings in both the parallel and recurrent formulations of retention.
    See: https://arxiv.org/pdf/2307.08621v3.pdf, Section 2.1 (Retention)

    NOTE: The actual values for thetas are not specified in the paper, so I
    copied these values from the official implementation.
    See: https://github.com/microsoft/torchscale/blob/7d231743f4f96c460b7cf0aa0cf242bb192b34f8/torchscale/architecture/retnet.py#L27C1-L28C59
    """
    x = torch.linspace(0, 1, steps=head_dim // 2, device=device, dtype=dtype)
    thetas = 1 / (scale**x)
    return repeat(thetas, "d -> (d n)", n=2)


# NOTE: For the purposes of positional embeddings, we view query/key Tensors as
# complex-valued, where the even-numbered indices are the real part, and the
# odd-numbered indices are the imaginary part.  This makes it easy to compute
# complex values without *actually* using complex dtypes in PyTorch.
# (Complex dtypes have limited support compared to real dtypes.)
#
# I don't re-explain this in the functions below, but it's important to keep in
# mind when reading the code.


def _multiply_by_i(x: Tensor) -> Tensor:
    """Multiply a complex-valued tensor by the imaginary unit 'i'."""
    return torch.stack((-x[..., 1::2], x[..., ::2]), dim=-1).flatten(start_dim=-2)


@torch.jit.script
def _theta_shift(x: Tensor, sin: Tensor, cos: Tensor) -> Tensor:
    # TODO: Add docstring
    return (x * cos) + (_multiply_by_i(x) * sin)


def retention_parallel(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: Optional[float] = None,
    need_weights: bool = False,
) -> Tuple[Tensor, Optional[Tensor]]:
    decay_mask = _build_decay_mask(
        num_heads=query.shape[1],
        query_length=query.shape[2],
        key_length=key.shape[2],
        device=query.device,
        dtype=query.dtype,
    )

    # einstein notation:
    # - b: batch_size
    # - h: num_heads
    # - n / s: seq_length
    # - d: hidden_dim
    if scale is None:
        scale = key.size(-1) ** 0.5
    key = key / scale

    similarity = einsum(query, key, "b h n d, b h s d -> b h n s")
    similarity = similarity * rearrange(decay_mask, "h n s -> () h n s")
    retention = einsum(similarity, value, "b h n s, b h s d -> b h n d")

    if need_weights:
        return retention, similarity
    else:
        return retention, None


def retention_recurrent(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    prev_state: Optional[Tensor],
    scale: Optional[float] = None,
) -> Tuple[Tensor, Tensor]:
    decay_gammas = _build_decay_gammas(
        num_heads=query.shape[1], device=query.device, dtype=query.dtype
    )

    # einstein notation:
    # - b: batch_size
    # - h: num_heads
    # - d: hidden_dim
    if scale is None:
        scale = key.size(-1) ** 0.5
    key = key / scale

    state = einsum(key, value, "b h d, b h m -> b h d m")
    if prev_state is not None:
        state = state + prev_state * rearrange(decay_gammas, "h -> () h () ()")
    retention = einsum(query, state, "b h d, b h d m -> b h m")

    return retention, state


def retention_chunkwise(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    prev_state: Optional[Tensor],
    scale: Optional[float] = None,
) -> Tuple[Tensor, Tensor]:
    decay_gammas = _build_decay_gammas(
        num_heads=query.shape[1], device=query.device, dtype=query.dtype
    )
    decay_mask = _build_decay_mask(
        num_heads=query.shape[1],
        query_length=query.shape[2],
        key_length=key.shape[2],
        device=query.device,
        dtype=query.dtype,
    )

    # einstein notation:
    # - b: batch_size
    # - h: num_heads
    # - n / s: seq_length
    # - d: head_dim
    if scale is None:
        scale = key.size(-1) ** 0.5
    key = key / scale

    # intra-chunk (same as parallel retention)
    similarity = einsum(query, key, "b h n d, b h s d -> b h n s")
    similarity = similarity * rearrange(decay_mask, "h n s -> () h n s")
    retention = einsum(similarity, value, "b h n s, b h s d -> b h n d")

    # cross-chunk (derived from recurrent retention)
    decay_gammas = rearrange(decay_gammas, "h -> () h () ()")
    inner_pos = rearrange(
        torch.arange(key.size(2), device=key.device, dtype=key.dtype) + 1,
        "n -> () () n ()",
    )
    state_decays = decay_gammas ** (key.size(2) - inner_pos)
    discounted_key = einsum(key, state_decays, 'b h n d, _ h n _ -> b h n d')
    state = einsum(discounted_key, value, 'b h n d1, b h n d2 -> b h d1 d2')
    if prev_state is not None:
        # Update internal state to return to the user
        chunk_decay = decay_gammas ** key.size(2)
        state = state + prev_state * chunk_decay
        # Update the retention Tensor, based on cross-chunk information
        inner_decay = decay_gammas**inner_pos
        retention = retention + (
            einsum(query, prev_state, "b h n d1, b h d1 d2 -> b h n d2") * inner_decay
        )

    return retention, state


class MultiScaleRetention(nn.Module):
    """Multi-scale retention (MSR) layer.  Intended to be (mostly) a drop-in replacement
    for nn.MultiheadAttention, but with the option to use either the parallel or
    recurrent formulation of retention. (Attention only has the parallel formulation.)

    NOTE: As presented in the paper, Multi-Scale Retention includes an explicit
    position embedding, which is based on xPos.  IMO, this is unnecessary and overly
    specific to language modeling, since other domains (e.g. computer vision,
    heterogeneous graphs) will have different positional semantics.

    I have made the relational position embedding optional, so that this module
    can (in theory) support more modalities. Setting 'relative_position=False' will
    remove the positional embedding, and instead rely on the query and key
    embeddings to encode positional information ahead of time (if needed at all).
    See: https://github.com/microsoft/torchscale/issues/48

    Reference:
        "Retentive Network: A Successor to Transformer for Large Language Models"
        https://arxiv.org/pdf/2307.08621v3.pdf
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout: float = 0.0,
        relative_position: bool = True,
        bias: bool = True,
        batch_first: bool = True,
        activation: Union[ActivationString, Callable[[Tensor], Tensor]] = "swish",
        group_norm_eps: float = 1e-6,
        device: Optional[Union[torch.device, str]] = None,
        dtype: Optional[torch.dtype] = None,
        # TODO???
        # add_bias_kv=False,
        # add_zero_attn=False,
        # kdim=None,
        # vdim=None,
    ):
        if not batch_first:
            raise NotImplementedError("batch_first=False is not yet supported")
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)

        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.relative_position = relative_position
        self.bias = bias
        self.activation = activation

        if embed_dim % self.num_heads != 0:
            raise ValueError(
                f"embed_dim ({embed_dim}) must be divisible by num_heads ({num_heads})"
            )

        head_dim = embed_dim // num_heads
        if not head_dim % 8 == 0:
            raise ValueError(
                f"head_dim (embed_dim / num_heads = {head_dim}) must be divisible by 8"
            )

        # The q/k/v projection layers are the same as in vanilla MHA.
        self.q_proj = nn.Linear(
            embed_dim, embed_dim, bias=bias, device=device, dtype=dtype
        )
        self.k_proj = nn.Linear(
            embed_dim, embed_dim, bias=bias, device=device, dtype=dtype
        )
        self.v_proj = nn.Linear(
            embed_dim, embed_dim, bias=bias, device=device, dtype=dtype
        )
        self.group_norm = nn.GroupNorm(
            num_groups=num_heads,
            num_channels=embed_dim,
            affine=False,
            eps=group_norm_eps,
            device=device,
            dtype=dtype,
        )
        # The output project is slightly different, due to the gated "swish" layer.
        self.g_proj = nn.Linear(
            embed_dim, embed_dim, bias=bias, device=device, dtype=dtype
        )
        self.out_proj = nn.Linear(
            embed_dim, embed_dim, bias=bias, device=device, dtype=dtype
        )

        # 'thetas' parameter for updating the relative position embeddings.
        thetas: Optional[Tensor] = None
        if relative_position:
            thetas = _build_position_thetas(
                head_dim=head_dim, device=device, dtype=dtype
            )
        self.thetas: Optional[Tensor]
        self.register_buffer("thetas", thetas)

        self._reset_parameters()

    def _reset_parameters(self):
        # TODO: Double-check that we're following the same initialization as in
        # the paper.  This is a generic initialization for MHA linear layers.
        nn.init.xavier_normal_(self.q_proj.weight)
        if self.q_proj.bias is not None:
            nn.init.constant_(self.q_proj.bias, 0)
        nn.init.xavier_normal_(self.k_proj.weight)
        if self.k_proj.bias is not None:
            nn.init.constant_(self.k_proj.bias, 0)
        nn.init.xavier_normal_(self.v_proj.weight)
        if self.v_proj.bias is not None:
            nn.init.constant_(self.v_proj.bias, 0)
        nn.init.xavier_normal_(self.out_proj.weight)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0)
        nn.init.xavier_normal_(self.g_proj.weight)
        if self.g_proj.bias is not None:
            nn.init.constant_(self.g_proj.bias, 0)

    def forward_parallel(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        need_weights: bool = False,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        # einstein notation:
        # b - batch size
        # n - sequence length
        # h - number of heads
        # d - embedding dimension
        #
        # Input shape: (b, n, d)
        q: Tensor = self.q_proj(query)
        k: Tensor = self.k_proj(key)
        v: Tensor = self.v_proj(value)

        # Unfold 'd' dimension into 'h' separate retention heads.  Move the head
        # dimension to position 1 (makes matrix ops *much* faster).
        q = rearrange(q, "b n (h d) -> b h n d", h=self.num_heads)
        k = rearrange(k, "b n (h d) -> b h n d", h=self.num_heads)
        v = rearrange(v, "b n (h d) -> b h n d", h=self.num_heads)

        if self.relative_position:
            assert self.thetas is not None
            indices = torch.arange(q.size(2), device=q.device, dtype=q.dtype)
            indices = rearrange(indices, "n -> () () n ()")
            thetas = rearrange(self.thetas, "d -> () () () d")
            angles = indices * thetas
            sin = torch.sin(angles)
            cos = torch.cos(angles)

            q = _theta_shift(q, sin, cos)
            k = _theta_shift(k, sin, cos)

        # Apply retention then group norm.
        retention, weights = retention_parallel(q, k, v, need_weights=need_weights)
        # To apply group norm in an equivalent way to the recurrent formulation,
        # we fold the sequence dimension into the batch dimension.  Otherwise,
        # normalization would be applied over the entire input sequence.
        batch_size = retention.size(0)
        retention = rearrange(retention, "b h n d -> (b n) (h d)")
        retention = F.dropout(retention, p=self.dropout, training=self.training)
        retention = self.group_norm(retention)
        # Unfold 'n' from the batch dimension, and fold 'h' back into the embed dim.
        retention = rearrange(retention, "(b n) e -> b n e", b=batch_size)

        # NOTE: Unlike multihead attention, the retention paper applies a "swish"
        # gate to increase the non-linear capacity of the model.  (IMO this is likely
        # to make up for the lack of "softmax" activation in the retention mechanism.)
        #
        # The paper describes the gate as:
        #   g = swish(X * W_g)
        # where X is the input to the layer.  The authors use Retention in a
        # Decoder-only model, the q/k/v inputs are the same (i.e. X = q = k = v).
        # So, I assume that 'query' can equivalently be used as the input.
        gate = self.activation(self.g_proj(query))
        retention = self.out_proj(retention * gate)

        return retention, weights

    def forward_recurrent(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        seq_idx: int,
        prev_state: Optional[Tensor],
    ) -> Tuple[Tensor, Tensor]:
        # einstein notation:
        # b - batch size
        # h - number of heads
        # d - embedding dimension
        #
        # input shape: (b, d)
        q: Tensor = self.q_proj(query)
        k: Tensor = self.k_proj(key)
        v: Tensor = self.v_proj(value)

        # Unfold 'd' dimension into 'h' separate retention heads.
        q = rearrange(q, "b (h d) -> b h d", h=self.num_heads)
        k = rearrange(k, "b (h d) -> b h d", h=self.num_heads)
        v = rearrange(v, "b (h d) -> b h d", h=self.num_heads)

        if self.relative_position:
            assert self.thetas is not None
            thetas = rearrange(self.thetas, "d -> () () d")
            angles = seq_idx * thetas
            sin = torch.sin(angles)
            cos = torch.cos(angles)

            q = _theta_shift(q, sin, cos)
            k = _theta_shift(k, sin, cos)

        # Apply retention then group norm.
        retention, state = retention_recurrent(q, k, v, prev_state=prev_state)
        retention = F.dropout(retention, p=self.dropout, training=self.training)
        # Fold heads back into the embedding dimension.
        retention = rearrange(retention, "b h d -> b (h d)")
        retention = self.group_norm(retention)

        # NOTE: Unlike multihead attention, the retention paper applies a "swish"
        # gate to increase the non-linear capacity of the model.  (IMO this is likely
        # to make up for the lack of "softmax" activation in the retention mechanism.)
        #
        # The paper describes the gate as:
        #   g = swish(X * W_g)
        # where X is the input to the layer.  The authors use Retention in a
        # Decoder-only model, the q/k/v inputs are the same (i.e. X = q = k = v).
        # So, I assume that 'query' can equivalently be used as the input.
        gate = self.activation(self.g_proj(query))
        retention = self.out_proj(retention * gate)

        return retention, state

    def forward_chunkwise(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        start_idx: int,
        prev_state: Optional[Tensor],
    ) -> Tuple[Tensor, Tensor]:
        # einstein notation:
        # b - batch size
        # n - sequence length
        # h - number of heads
        # d - embedding dimension
        #
        # Input shape: (b, n, d)
        q: Tensor = self.q_proj(query)
        k: Tensor = self.k_proj(key)
        v: Tensor = self.v_proj(value)

        # Unfold 'd' dimension into 'h' separate retention heads.  Move the head
        # dimension to position 1 (makes matrix ops *much* faster).
        q = rearrange(q, "b n (h d) -> b h n d", h=self.num_heads)
        k = rearrange(k, "b n (h d) -> b h n d", h=self.num_heads)
        v = rearrange(v, "b n (h d) -> b h n d", h=self.num_heads)

        if self.relative_position:
            # global (cross-chunk) + intra-chunk relative position embedding
            assert self.thetas is not None
            indices = torch.arange(start_idx, start_idx + q.size(2), device=q.device, dtype=q.dtype)
            indices = rearrange(indices, "n -> () () n ()")
            thetas = rearrange(self.thetas, "d -> () () () d")
            angles = indices * thetas
            sin = torch.sin(angles)
            cos = torch.cos(angles)
            q = _theta_shift(q, sin, cos)
            k = _theta_shift(k, sin, cos)

        # Apply retention then group norm.
        retention, state = retention_chunkwise(q, k, v, prev_state=prev_state)
        # To apply group norm in an equivalent way to the recurrent formulation,
        # we fold the sequence dimension into the batch dimension.  Otherwise,
        # normalization would be applied over the entire input sequence.
        batch_size = retention.size(0)
        retention = rearrange(retention, "b h n d -> (b n) (h d)")
        retention = F.dropout(retention, p=self.dropout, training=self.training)
        retention = self.group_norm(retention)
        # Unfold 'n' from the batch dimension, and fold 'h' back into the embed dim.
        retention = rearrange(retention, "(b n) e -> b n e", b=batch_size)

        # NOTE: Unlike multihead attention, the retention paper applies a "swish"
        # gate to increase the non-linear capacity of the model.  (IMO this is likely
        # to make up for the lack of "softmax" activation in the retention mechanism.)
        #
        # The paper describes the gate as:
        #   g = swish(X * W_g)
        # where X is the input to the layer.  The authors use Retention in a
        # Decoder-only model, the q/k/v inputs are the same (i.e. X = q = k = v).
        # So, I assume that 'query' can equivalently be used as the input.
        gate = self.activation(self.g_proj(query))
        retention = self.out_proj(retention * gate)

        return retention, state

    def forward(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        need_weights: bool = False,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        return self.forward_parallel(query, key, value, need_weights=need_weights)


if __name__ == "__main__":
    batch_size = 1
    seq_len = 3
    embed_dim = 16
    num_heads = 2

    layer = MultiScaleRetention(
        embed_dim=embed_dim,
        num_heads=num_heads,
    )

    x = torch.randn(batch_size, seq_len, embed_dim)
    y_chunkwise, chunkwise_state = retention_chunkwise(x, x, x, prev_state=None)
    print(chunkwise_state)
    print("-" * 40)

    y_recurrent = torch.zeros_like(y_chunkwise)
    recurrent_state = None
    for i in range(seq_len):
        y_recurrent[:, :, i, :], recurrent_state = retention_recurrent(
            x[:, :, i, :], x[:, :, i, :], x[:, :, i, :], prev_state=recurrent_state
        )
    print(recurrent_state)



import os
from subprocess import getoutput


def get_version_tag() -> str:
    try:
        env_key = "YET_ANOTHER_RETNET_VERSION".upper()
        version = os.environ[env_key]
    except KeyError:
        version = getoutput("git describe --tags --abbrev=0")

    if version.lower().startswith("fatal"):
        version = "0.0.0"

    return version


VERSION = get_version_tag()



from math import ceil
from timeit import Timer
from typing import Callable, List, NamedTuple

import torch


class BenchmarkResult(NamedTuple):
    mean: float
    std: float

    def __repr__(self):
        return f"BenchmarkResult(mean: {self.mean:.3e}, std: {self.std:.3e})"

    def __str__(self):
        return f"({self.mean:.3e} \u00B1 {self.std:.3e}) s"


@torch.no_grad()
def benchmark(
    fn: Callable,
    *args,
    min_total_seconds: float = 1.0,
    min_iterations: int = 10,
    **kwargs,
) -> BenchmarkResult:
    # Benchmark the runtime of a function and dynamically determine the number of
    # iterations to run.  Continue running the function until *total* runtime
    # exceeds 'min_total_seconds' and 'min_iterations'.
    if min_iterations < 2:
        raise ValueError("min_iterations must be >= 2")

    timer = Timer(
        "fn(*args, **kwargs); synchronize()",
        globals={
            "fn": fn,
            "args": args,
            "kwargs": kwargs,
            "synchronize": (
                torch.cuda.synchronize
                if torch.cuda.is_available()
                else (lambda: None)  # type: ignore
            ),
        },
    )
    # Run the function 5 times to warm up
    _ = timer.repeat(number=1, repeat=5)

    times: List[float] = []
    total_time = 0.0
    num_iterations = min_iterations

    while total_time < min_total_seconds:
        _times = timer.repeat(number=1, repeat=num_iterations)
        times.extend(_times)

        times_tensor = torch.as_tensor(times)
        total_time = times_tensor.sum().item()
        avg_time = times_tensor.mean().item()
        num_iterations = ceil((min_total_seconds - total_time) / avg_time)

    times_tensor = torch.as_tensor(times)
    return BenchmarkResult(
        mean=times_tensor.mean().item(),
        std=times_tensor.std().item(),
    )



import random
from typing import Generator, List, Literal, Optional

import requests
from torchdata.datapipes.iter import IterableWrapper, IterDataPipe

GUTENBERG_TOP_100_URLS = [
    # Romeo and Juliet by William Shakespeare
    "https://www.gutenberg.org/ebooks/1513.txt.utf-8",
    # Moby Dick; Or The Whale by Herman Melville
    "https://www.gutenberg.org/ebooks/2701.txt.utf-8",
    # A Room with a View by E. M. Forster
    "https://www.gutenberg.org/ebooks/2641.txt.utf-8",
    # Middlemarch by George Eliot
    "https://www.gutenberg.org/ebooks/145.txt.utf-8",
    # Little Women; Or Meg Jo Beth and Amy by Louisa May Alcott
    "https://www.gutenberg.org/ebooks/37106.txt.utf-8",
    # The Complete Works of William Shakespeare by William Shakespeare
    "https://www.gutenberg.org/ebooks/100.txt.utf-8",
    # The Blue Castle: a novel by L. M. Montgomery
    "https://www.gutenberg.org/ebooks/67979.txt.utf-8",
    # The Enchanted April by Elizabeth Von Arnim
    "https://www.gutenberg.org/ebooks/16389.txt.utf-8",
    # The Adventures of Ferdinand Count Fathom — Complete by T. Smollett
    "https://www.gutenberg.org/ebooks/6761.txt.utf-8",
    # Cranford by Elizabeth Cleghorn Gaskell
    "https://www.gutenberg.org/ebooks/394.txt.utf-8",
    # The Expedition of Humphry Clinker by T. Smollett
    "https://www.gutenberg.org/ebooks/2160.txt.utf-8",
    # The Adventures of Roderick Random by T. Smollett
    "https://www.gutenberg.org/ebooks/4085.txt.utf-8",
    # History of Tom Jones a Foundling by Henry Fielding
    "https://www.gutenberg.org/ebooks/6593.txt.utf-8",
    # Twenty Years After by Alexandre Dumas
    "https://www.gutenberg.org/ebooks/1259.txt.utf-8",
    # My Life — Volume 1 by Richard Wagner
    "https://www.gutenberg.org/ebooks/5197.txt.utf-8",
    # Pride and Prejudice by Jane Austen
    "https://www.gutenberg.org/ebooks/1342.txt.utf-8",
    # Down and Out in the Magic Kingdom by Cory Doctorow
    "https://www.gutenberg.org/ebooks/8086.txt.utf-8",
    # The Odyssey by Homer
    "https://www.gutenberg.org/ebooks/1727.txt.utf-8",
    # Alice's Adventures in Wonderland by Lewis Carroll
    "https://www.gutenberg.org/ebooks/11.txt.utf-8",
    # Frankenstein; Or The Modern Prometheus by Mary Wollstonecraft Shelley
    "https://www.gutenberg.org/ebooks/84.txt.utf-8",
    # Anna Karenina by graf Leo Tolstoy
    "https://www.gutenberg.org/ebooks/1399.txt.utf-8",
    # A Tale of Two Cities by Charles Dickens
    "https://www.gutenberg.org/ebooks/98.txt.utf-8",
    # Dracula by Bram Stoker
    "https://www.gutenberg.org/ebooks/345.txt.utf-8",
    # The Adventures of Sherlock Holmes by Arthur Conan Doyle
    "https://www.gutenberg.org/ebooks/1661.txt.utf-8",
    # The Picture of Dorian Gray by Oscar Wilde
    "https://www.gutenberg.org/ebooks/174.txt.utf-8",
    # The Count of Monte Cristo Illustrated by Alexandre Dumas
    "https://www.gutenberg.org/ebooks/1184.txt.utf-8",
    # The Brothers Karamazov by Fyodor Dostoyevsky
    "https://www.gutenberg.org/ebooks/28054.txt.utf-8",
    # War and Peace by graf Leo Tolstoy
    "https://www.gutenberg.org/ebooks/2600.txt.utf-8",
    # How to Pick a Mate: The Guide to a Happy Marriage by Clifford R. Adams and Vance Packard
    "https://www.gutenberg.org/ebooks/67472.txt.utf-8",
    # The Great Gatsby by F. Scott Fitzgerald
    "https://www.gutenberg.org/ebooks/64317.txt.utf-8",
    # Metamorphosis by Franz Kafka
    "https://www.gutenberg.org/ebooks/5200.txt.utf-8",
    # Ulysses by James Joyce
    "https://www.gutenberg.org/ebooks/4300.txt.utf-8",
    # The Prince by Niccolò Machiavelli
    "https://www.gutenberg.org/ebooks/1232.txt.utf-8",
    # Crime and Punishment by Fyodor Dostoyevsky
    "https://www.gutenberg.org/ebooks/2554.txt.utf-8",
    # The Yellow Wallpaper by Charlotte Perkins Gilman
    "https://www.gutenberg.org/ebooks/1952.txt.utf-8",
    # The Romance of Lust: A classic Victorian erotic novel by Anonymous
    "https://www.gutenberg.org/ebooks/30254.txt.utf-8",
    # The Kama Sutra of Vatsyayana by Vatsyayana
    "https://www.gutenberg.org/ebooks/27827.txt.utf-8",
    # Grimms' Fairy Tales by Jacob Grimm and Wilhelm Grimm
    "https://www.gutenberg.org/ebooks/2591.txt.utf-8",
    # A Modest Proposal by Jonathan Swift
    "https://www.gutenberg.org/ebooks/1080.txt.utf-8",
    # The Iliad by Homer
    "https://www.gutenberg.org/ebooks/6130.txt.utf-8",
    # Great Expectations by Charles Dickens
    "https://www.gutenberg.org/ebooks/1400.txt.utf-8",
    # Thus Spake Zarathustra: A Book for All and None by Friedrich Wilhelm Nietzsche
    "https://www.gutenberg.org/ebooks/1998.txt.utf-8",
    # The Adventures of Tom Sawyer Complete by Mark Twain
    "https://www.gutenberg.org/ebooks/74.txt.utf-8",
    # Meditations by Emperor of Rome Marcus Aurelius
    "https://www.gutenberg.org/ebooks/2680.txt.utf-8",
    # The Wonderful Wizard of Oz by L. Frank Baum
    "https://www.gutenberg.org/ebooks/55.txt.utf-8",
    # A Doll's House : a play by Henrik Ibsen
    "https://www.gutenberg.org/ebooks/2542.txt.utf-8",
    # The Importance of Being Earnest: A Trivial Comedy for Serious People by Oscar Wilde
    "https://www.gutenberg.org/ebooks/844.txt.utf-8",
    # Moby Multiple Language Lists of Common Words by Grady Ward
    "https://www.gutenberg.org/ebooks/3206.txt.utf-8",
    # Adventures of Huckleberry Finn by Mark Twain
    "https://www.gutenberg.org/ebooks/76.txt.utf-8",
    # The slang dictionary : by John Camden Hotten
    "https://www.gutenberg.org/ebooks/42108.txt.utf-8",
    # Jane Eyre: An Autobiography by Charlotte Brontë
    "https://www.gutenberg.org/ebooks/1260.txt.utf-8",
    # The Prophet by Kahlil Gibran
    "https://www.gutenberg.org/ebooks/58585.txt.utf-8",
    # Beyond Good and Evil by Friedrich Wilhelm Nietzsche
    "https://www.gutenberg.org/ebooks/4363.txt.utf-8",
    # Anne of Green Gables by L. M. Montgomery
    "https://www.gutenberg.org/ebooks/45.txt.utf-8",
    # On the Duty of Civil Disobedience by Henry David Thoreau
    "https://www.gutenberg.org/ebooks/71.txt.utf-8",
    # Tractatus Logico-Philosophicus by Ludwig Wittgenstein
    "https://www.gutenberg.org/ebooks/5740.txt.utf-8",
    # The Strange Case of Dr. Jekyll and Mr. Hyde by Robert Louis Stevenson
    "https://www.gutenberg.org/ebooks/43.txt.utf-8",
    # Don Quixote by Miguel de Cervantes Saavedra
    "https://www.gutenberg.org/ebooks/996.txt.utf-8",
    # Calculus Made Easy by Silvanus P. Thompson
    "https://www.gutenberg.org/ebooks/33283.txt.utf-8",
    # Mark Twain's Speeches by Mark Twain
    "https://www.gutenberg.org/ebooks/3188.txt.utf-8",
    # Treasure Island by Robert Louis Stevenson
    "https://www.gutenberg.org/ebooks/120.txt.utf-8",
    # The Rámáyan of Válmíki translated into English verse by Valmiki
    "https://www.gutenberg.org/ebooks/24869.txt.utf-8",
    # A Child's History of the World by V. M. Hillyer
    "https://www.gutenberg.org/ebooks/67149.txt.utf-8",
    # A Study in Scarlet by Arthur Conan Doyle
    "https://www.gutenberg.org/ebooks/244.txt.utf-8",
    # Walden and On The Duty Of Civil Disobedience by Henry David Thoreau
    "https://www.gutenberg.org/ebooks/205.txt.utf-8",
    # Wuthering Heights by Emily Brontë
    "https://www.gutenberg.org/ebooks/768.txt.utf-8",
    # The Republic by Plato
    "https://www.gutenberg.org/ebooks/1497.txt.utf-8",
    # Essays of Michel de Montaigne — Complete by Michel de Montaigne
    "https://www.gutenberg.org/ebooks/3600.txt.utf-8",
    # Winnie-the-Pooh by A. A. Milne
    "https://www.gutenberg.org/ebooks/67098.txt.utf-8",
    # Emma by Jane Austen
    "https://www.gutenberg.org/ebooks/158.txt.utf-8",
    # The divine comedy by Dante Alighieri
    "https://www.gutenberg.org/ebooks/8800.txt.utf-8",
    # The Scarlet Letter by Nathaniel Hawthorne
    "https://www.gutenberg.org/ebooks/25344.txt.utf-8",
    # Little Women by Louisa May Alcott
    "https://www.gutenberg.org/ebooks/514.txt.utf-8",
    # Les Misérables by Victor Hugo
    "https://www.gutenberg.org/ebooks/135.txt.utf-8",
    # Demonology and Devil-lore by Moncure Daniel Conway
    "https://www.gutenberg.org/ebooks/40686.txt.utf-8",
    # Spoon River Anthology by Edgar Lee Masters
    "https://www.gutenberg.org/ebooks/1280.txt.utf-8",
    # The King James Version of the Bible
    "https://www.gutenberg.org/ebooks/10.txt.utf-8",
    # Peter Pan by J. M. Barrie
    "https://www.gutenberg.org/ebooks/16.txt.utf-8",
    # The King in Yellow by Robert W. Chambers
    "https://www.gutenberg.org/ebooks/8492.txt.utf-8",
    # The War of the Worlds by H. G. Wells
    "https://www.gutenberg.org/ebooks/36.txt.utf-8",
    # Notes from the Underground by Fyodor Dostoyevsky
    "https://www.gutenberg.org/ebooks/600.txt.utf-8",
    # The Problems of Philosophy by Bertrand Russell
    "https://www.gutenberg.org/ebooks/5827.txt.utf-8",
    # Heart of Darkness by Joseph Conrad
    "https://www.gutenberg.org/ebooks/219.txt.utf-8",
    # The Time Machine by H. G. Wells
    "https://www.gutenberg.org/ebooks/35.txt.utf-8",
    # A Christmas Carol in Prose; Being a Ghost Story of Christmas by Charles Dickens
    "https://www.gutenberg.org/ebooks/46.txt.utf-8",
    # Carmilla by Joseph Sheridan Le Fanu
    "https://www.gutenberg.org/ebooks/10007.txt.utf-8",
    # Dubliners by James Joyce
    "https://www.gutenberg.org/ebooks/2814.txt.utf-8",
    # David Copperfield by Charles Dickens
    "https://www.gutenberg.org/ebooks/766.txt.utf-8",
    # Sense and Sensibility by Jane Austen
    "https://www.gutenberg.org/ebooks/161.txt.utf-8",
    # Dombey and Son by Charles Dickens
    "https://www.gutenberg.org/ebooks/821.txt.utf-8",
    # The Jungle Book by Rudyard Kipling
    "https://www.gutenberg.org/ebooks/236.txt.utf-8",
    # The Art of War by active 6th century B.C. Sunzi
    "https://www.gutenberg.org/ebooks/132.txt.utf-8",
    # The murder of Roger Ackroyd by Agatha Christie
    "https://www.gutenberg.org/ebooks/69087.txt.utf-8",
    # The Souls of Black Folk by W. E. B. Du Bois
    "https://www.gutenberg.org/ebooks/408.txt.utf-8",
    # The Marching Morons by C. M. Kornbluth
    "https://www.gutenberg.org/ebooks/51233.txt.utf-8",
    # Josefine Mutzenbacher by Felix Salten
    "https://www.gutenberg.org/ebooks/31284.txt.utf-8",
    # Oliver Twist by Charles Dickens
    "https://www.gutenberg.org/ebooks/730.txt.utf-8",
    # Through the Looking-Glass by Lewis Carroll
    "https://www.gutenberg.org/ebooks/12.txt.utf-8",
    # The Confessions of St. Augustine by Bishop of Hippo Saint Augustine
    "https://www.gutenberg.org/ebooks/3296.txt.utf-8",
    # Struwwelpeter: Merry Stories and Funny Pictures by Heinrich Hoffmann
    "https://www.gutenberg.org/ebooks/12116.txt.utf-8",
]


def get_split_indices(
    num_samples: int,
    split: Literal["train", "val", "test"],
    seed: int = 42,
    val_split: float = 0.1,
    test_split: float = 0.1,
) -> List[int]:
    indices = list(range(num_samples))
    random.seed(seed)
    random.shuffle(indices)

    num_val = int(num_samples * val_split)
    num_test = int(num_samples * test_split)

    if split == "train":
        out = indices[num_val + num_test :]
    elif split == "val":
        out = indices[:num_val]
    elif split == "test":
        out = indices[num_val : num_val + num_test]
    else:
        raise ValueError(f"Invalid split: {split}")

    return sorted(out)


class GutenbergEBookLoader(IterDataPipe[str]):
    def __init__(self, dp: IterDataPipe[str]):
        self.dp = dp

    def __iter__(self) -> Generator[str, None, None]:
        for url in self.dp:
            resp = requests.get(url)
            try:
                resp.raise_for_status()
            except requests.exceptions.HTTPError:
                continue

            text = resp.text
            text = text.split("***", maxsplit=2)[-1]  # remove header
            text = text.split("*** End", maxsplit=1)[0]  # remove footer
            text = text.strip()

            yield text


class TextChunker(IterDataPipe[str]):
    def __init__(
        self,
        dp: IterDataPipe[str],
        chunk_size: int = 4096,
        step_size: Optional[int] = None,
        drop_last: bool = False,
    ):
        self.dp = dp
        self.chunk_size = chunk_size
        self.step_size = step_size or chunk_size
        self.drop_last = drop_last

    def __iter__(self) -> Generator[str, None, None]:
        for text in self.dp:
            for i in range(0, len(text), self.step_size):
                chunk = text[i : i + self.chunk_size]
                if self.drop_last and len(chunk) < self.chunk_size:
                    continue

                chunk = chunk.split(" ", maxsplit=1)[-1]  # leading partial words
                chunk = chunk.rsplit(" ", maxsplit=1)[0]  # trailing partial words
                chunk = chunk.strip()  # leading/trailing whitespace
                yield chunk


def project_gutenberg_top_100_datapipe(
    split: Literal["train", "val", "test"],
    chunk_size: int = 4096,
    step_size: Optional[int] = None,
    shuffle: bool = False,
    shuffle_buffer_size: int = 8192,
    drop_last: bool = True,
) -> IterDataPipe[str]:
    indices = get_split_indices(len(GUTENBERG_TOP_100_URLS), split=split)
    if shuffle:
        random.shuffle(indices)

    # Iterable datapipe of ebook URLs (ebooks are UTF-8 text files)
    pipe: IterDataPipe = IterableWrapper([GUTENBERG_TOP_100_URLS[i] for i in indices])
    pipe = GutenbergEBookLoader(pipe)
    pipe = TextChunker(
        pipe, chunk_size=chunk_size, step_size=step_size, drop_last=drop_last
    )
    pipe = pipe.sharding_filter()
    if shuffle:
        pipe = pipe.shuffle(buffer_size=shuffle_buffer_size)

    return pipe



from typing import Optional

import pytest
import torch
from torch import Tensor

from yet_another_retnet.retention import (
    MultiScaleRetention,
    retention_chunkwise,
    retention_parallel,
    retention_recurrent,
)

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float64
# Set deterministic CUDA ops
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False


def test_retention_parallel_forward():
    # TODO
    pass


def test_retention_recurrent_forward():
    # TODO
    pass


def test_retention_chunkwise_forward():
    # TODO
    pass


@torch.no_grad()
@pytest.mark.parametrize("batch_size", [2])
@pytest.mark.parametrize("num_heads", [2, 4])
@pytest.mark.parametrize("seq_length", [16])
@pytest.mark.parametrize("hidden_dim", [4, 8])
@pytest.mark.parametrize("chunk_size", [1, 4, 16])
def test_equivalent_retention_formulations(
    batch_size: int,
    num_heads: int,
    seq_length: int,
    hidden_dim: int,
    chunk_size: int,
):
    size = (batch_size, num_heads, seq_length, hidden_dim)
    query = torch.randn(*size, device=DEVICE, dtype=DTYPE)
    key = torch.randn(*size, device=DEVICE, dtype=DTYPE)
    value = torch.randn(*size, device=DEVICE, dtype=DTYPE)

    # Parallel formulation
    y_parallel, _ = retention_parallel(query, key, value)

    # Recurrent formulation
    y_recurrent = torch.zeros_like(y_parallel)
    prev_state: Optional[Tensor] = None
    for i in range(seq_length):
        q, k, v = query[:, :, i], key[:, :, i], value[:, :, i]
        y_recurrent[:, :, i], prev_state = retention_recurrent(q, k, v, prev_state)

    torch.testing.assert_close(y_parallel, y_recurrent, rtol=1e-4, atol=1e-4)
    recurrent_state = prev_state

    # Chunkwise formulation
    y_chunkwise = torch.zeros_like(y_parallel)
    prev_state = None
    for i in range(0, seq_length, chunk_size):
        q = query[:, :, i : i + chunk_size]
        k = key[:, :, i : i + chunk_size]
        v = value[:, :, i : i + chunk_size]
        y_chunkwise[:, :, i : i + chunk_size], prev_state = retention_chunkwise(
            q, k, v, prev_state
        )

    torch.testing.assert_close(y_parallel, y_chunkwise, rtol=1e-4, atol=1e-4)
    torch.testing.assert_close(recurrent_state, prev_state, rtol=1e-4, atol=1e-4)


def test_multiscale_retention_forward_parallel():
    # TODO
    pass


def test_multiscale_retention_forward_recurrent():
    # TODO
    pass


@torch.no_grad()
@pytest.mark.parametrize("batch_size", [2])
@pytest.mark.parametrize("num_heads", [1, 2])
@pytest.mark.parametrize("seq_length", [8])
@pytest.mark.parametrize("embed_dim", [16, 32])
@pytest.mark.parametrize("chunk_size", [1, 4, 16])
def test_equivalent_multiscale_formulations(
    batch_size: int,
    num_heads: int,
    seq_length: int,
    embed_dim: int,
    chunk_size: int,
):
    size = (batch_size, seq_length, embed_dim)
    query = torch.randn(*size, device=DEVICE, dtype=DTYPE)
    key = torch.randn(*size, device=DEVICE, dtype=DTYPE)
    value = torch.randn(*size, device=DEVICE, dtype=DTYPE)
    mhr = MultiScaleRetention(embed_dim, num_heads, device=DEVICE, dtype=DTYPE).eval()

    y_parallel, _ = mhr.forward_parallel(query, key, value)

    y_recurrent = torch.zeros_like(y_parallel)
    prev_state: Optional[Tensor] = None
    for i in range(seq_length):
        q, k, v = query[:, i], key[:, i], value[:, i]
        y_recurrent[:, i], prev_state = mhr.forward_recurrent(
            q, k, v, seq_idx=i, prev_state=prev_state
        )
    recurrent_state = prev_state

    torch.testing.assert_close(y_parallel, y_recurrent, rtol=1e-4, atol=1e-4)

    y_chunkwise = torch.zeros_like(y_parallel)
    prev_state = None
    for i in range(0, seq_length, chunk_size):
        q = query[:, i : i + chunk_size]
        k = key[:, i : i + chunk_size]
        v = value[:, i : i + chunk_size]
        y_chunkwise[:, i : i + chunk_size], prev_state = mhr.forward_chunkwise(
            q, k, v, start_idx=i, prev_state=prev_state
        )

    torch.testing.assert_close(y_parallel, y_chunkwise, rtol=1e-4, atol=1e-4)
    torch.testing.assert_close(recurrent_state, prev_state, rtol=1e-4, atol=1e-4)



import asyncio
from typing import Callable, NamedTuple

import torch

KiB = 2**10
MiB = 2**20
GiB = 2**30


def _bytes_to_string(bytes: float) -> str:
    if bytes > GiB:
        return f"{bytes / GiB:.2f} GiB"
    elif bytes > MiB:
        return f"{bytes / MiB:.2f} MiB"
    elif bytes > KiB:
        return f"{bytes / KiB:.2f} KiB"
    else:
        return f"{bytes:.2f} Bytes"


class Profile(NamedTuple):
    name: str
    peak: float
    mean: float
    median: float

    def __str__(self, indent=""):
        return (
            f"{indent}{self.name}("
            f"peak: {_bytes_to_string(self.peak)}, "
            f"mean: {_bytes_to_string(self.mean)}, "
            f"median: {_bytes_to_string(self.median)})"
        )


async def _run_async(fn: Callable, *args, **kwargs):
    def synchronous_fn():
        torch.cuda.synchronize()
        _ = fn(*args, **kwargs)
        torch.cuda.synchronize()

    loop = asyncio.get_running_loop()
    await loop.run_in_executor(None, synchronous_fn)


async def _profile(fn: Callable, *args, interval: float = 1e-4, **kwargs):
    done = False
    memories = []

    async def _watch_cuda_memory(device=None):
        nonlocal done, memories
        while not done:
            memory = torch.cuda.memory_allocated()
            memories.append(memory)
            await asyncio.sleep(interval)

    cuda_task = asyncio.create_task(_watch_cuda_memory())
    main = asyncio.create_task(_run_async(fn, *args, **kwargs))
    done, _ = await main, cuda_task

    return memories


def profile(fn: Callable, *args, interval: float = 1e-4, **kwargs) -> Profile:
    torch.cuda.empty_cache()
    memories = torch.as_tensor(
        asyncio.run(_profile(fn, *args, interval=interval, **kwargs)),
        dtype=torch.float32,
    )
    return Profile(
        name="cuda",
        peak=torch.max(memories).item(),
        mean=torch.mean(memories).item(),
        median=torch.median(memories).item(),
    )
